' Ram driver

VAR

  long  cog

PUB Start(ramptr) : okay
  stop
  okay := cog := cognew(@entry, ramptr) + 1  

PUB stop

'' Stop ram driver - frees a cog

  if cog
    cogstop(cog~ - 1)   

DAT

'*******************************
'* Assembly language Ram driver *
'*******************************

' check the line number. If it has changed, then start reading the next line in and put @ screen
' if it has not changed then repeat the loop
' if it is the last line then restart the ram counter variable

' currentline counts up from 0 to tiles*16
' the video driver recycles the same 8*256 bytes in the screen buffer
' first, wait for currentline to equal zero and when it does, start reading in ram to @screen
' check the currentline. Shift left by 2 and mask-and with 1. 0-3 is 0, 4-7 is 1, 8-11 is 0, 12-15 is 1 etc

' Ram format. Each line is up to 512 pixels long. 256 is the standard video driver but over 300 does work
' so each line is 9 bits
' Then there are 4 lines in a group. Read these all in from ram in one subroutine call
' So that is 11 bits.
' And then these are in two groups of 4. So that is 12 bits.
' So - once 8 lines have been read in (12 bits = 4096 bytes), then change the latch
' There are 256/8 = 32 groups of 8 lines so that is 5 bits of the latch that are used
' and the 512k ram chip can store 4 frames (if needed)
' P0-7 is the data
' P8-P19 is the address for 8 lines (12 bits)
' P20 controls the latch enable
' P21 is /rd, P22 is /wr and P23 is /oe (chip select)
' ** There are pullups on P20 to P23 as the ram chip needs to retain data when transferring from spin to cog **




        org

entry
                        mov             t1,par
                        mov             status,t1
                        add             t1,#4
                        rdlong          _nextline,t1               ' line number in hub (updated by graphics cog) use with rdlong t1,_nextline
                        add             t1,#4
                        rdlong          screen_buf,t1              ' screen buffer location
                        add             t1,#4
                        rdlong          pixels_x,t1                 ' pixels per line
                        add             t1,#4
                        rdlong          serial_buffer,t1            ' serial buffer for data coming in

                        mov             ramline,screen_buf
                        add             ramline,topram              ' location of the temp 256 byte buffer that is rewritten each alternate line
                        mov             serial_buf_last,serial_buffer
                        add             serial_buf_last,#256        ' last entry in the serial buffer            
                        mov             outa,oewrrd                 ' oe/wr/rd/latch all high  %00000000_11110000_00000000_00000000         
                        mov             dira,dira_pins              ' enable these pins for output  %00000000_11111111_11111111_00000000

' in 'spaghetti' code with jumps all over the place - but in a way that maximises speed (this works better than djnz loops)
' reset everything
' wait for zero line
' label1
' read line of data
' add 2 to local line counter
' if locallinecounter>=224 then back to beginning
' add 1 to blockcounter
' if blockcounter=16 then reset blockcounter and increment latchcounter
' wait for current_line to equal localline (will be 2,4,6,8 etc)
' goto label1

' *** don't forget # after all jmp and call instructions!!! ****

main_loop_1             mov             sram_address,#0                 ' start reading ram at address #0
                        mov             localcounter,#0                 ' my local line counter
                        mov             blockcounter,#0                 ' block is 256 bytes, 16 blocks per latch
                        mov             latchcounter,#0                 ' block is 4096 bytes reset block counter
                        call            #selectlatch                    ' latch to zero
main_loop_2             mov             hubaddr,ramline                 ' setup for rdblock
                        mov             len,pixels_x                    ' setup for rdblock
main_loop_3             rdlong          current_line,_nextline          ' read the line that is being displayed
                        cmp             current_line,localcounter wz    ' equal to localcounter?
              if_nz     jmp             #main_loop_3                    ' loop until it is
                        call            #rdblock                        ' read in 256 bytes
                        add             sram_address,#256               ' next 256 bytes
                        add             localcounter,#2                 ' add 2 get ready for next line
                        cmp             localcounter,#224 wc            ' end of the screen = 224
'              if_nc     jmp             #main_loop_1                    ' restart everything
              if_nc     jmp             #process_serial   ' any new pixels?  
                        add             blockcounter,#1                 ' add 1 to block counter
                        cmp             blockcounter,#16  wz            ' time for new block?
              if_nz     jmp             #main_loop_2                    ' skip if still on current block
                        mov             sram_address,#0                 ' reset the ram address
                        mov             blockcounter,#0                 ' reset blockcounter
                        add             latchcounter,#1                 ' add 1 to latchcounter
                        call            #selectlatch                    ' latch to latchcounter
                        jmp             #main_loop_2


' 10k pullups on /oe, /rd and /wr and latch and don't put leds on these three pins

'Read a "len" byte block given by "sram_address" into hub at "hubaddr"  - Cluso99's ram driver
' preserves sram_address
rdblock               
                        mov             t1,sram_address         ' store sram address
                        shl             t1,#8                   ' shift left so in the right position
                        or              t1,oerd                 ' %00000000_01010000_00000000_00000000 ' /oe and /rd low 
                        ' outa pre-filled with the address shifted over by 8 and the /oe and /rd low
                        mov             outa,t1                        ' send it out
                        'nop                             ' cluso's driver has a nop but it seems ok without one    
rdloop                  mov             t2, ina               ' read byte from SRAM \ ignores upper bits
                        wrbyte          t2, hubaddr           ' copy byte to hub    /
                        add             hubaddr, #1             ' inc hub pointer
                        add             outa, #(1 << 8)         ' inc sram address
                        djnz            len, #rdloop            ' loop for xxx bytes
                        mov             outa,oewrrd             ' oe,wr and rd and latch all high
rdblock_ret             ret


' pass latchcounter
selectlatch             mov              dira,blockmask            ' enable different pins to reading, ie P0-7 are outputs now
                        mov              t1,latchcounter           ' get the latch counter
                        and              t1,#%01111111              'mask off high bit as 512k/4096/16 is 128
                        mov              outa,t1                    ' output blockcount byte
                        and              outa,latchlow              ' set latch low
                        'nop                                                       ' delay not needed
                        'nop
                        mov              dira,dira_pins             ' enable these pins for output  %00000000_11111111_11111111_00000000
                        mov              outa,oewrrd                ' set oe,wr,rd and latch high
selectlatch_ret         ret
                      
' Write a "len" byte block given by t3 in external ram from hub at t4  - Cluso99's ram driver 
wrblock
                        mov             dira,pin0to23            ' dira pins %00000000_11111111_11111111_11111111                   
wrloop                  mov             t1,t3                    ' store sram address in t1
                        shl             t1,#8                    ' shift left so in the right position
                        rdbyte          t2,t4                    ' get the byte
                        or              t1,oewr                  ' or with %00000000_00110000_00000000_00000000
                        or              t1,t2                    ' or with the hub address
                        mov             outa,t1                  ' send it out
                        nop                                      ' wait a little
                        nop
                        or              outa,oewrrd              ' oe,wr and rd all high
                        add             t4,#1                    ' add 1 to hub address
                        add             t3,#1                    ' add 1 to sram address
                        djnz            len, #wrloop             ' loop for xxx bytes
                        mov              dira,dira_pins          ' enable these pins for output  %00000000_11111111_11111111_00000000 
                        mov             outa,oewrrd              ' oe,wr and rd and latch all high
wrblock_ret             ret





' read serial data. The first byte is the row number, then 256 bytes
' if the last byte is zero then has been read, if not zero then read
' if the last byte is not zero, then process the data and then write a zero to the last byte
' if processing and the row is even then store to hub ram
' if the row is odd then store to external ram

process_serial          rdbyte   t1,serial_buf_last                 ' value in the last byte of the buffer
                        cmp      t1,#0 wz
                  if_z  jmp      #main_loop_1                       ' do nothing if zero
                        mov      t2,serial_buffer                   ' t2 is a counter
                        mov      t3,screen_buf                      ' t3 is the screen buffer
                        rdbyte   t1,t2                              ' t1 is the row number
                        mov      t5,t1                              ' t5 also contains the row number            
                        add      t2,#1                              ' increment - 1st byte is row number, data starts at 1
                        mov      t4,#256                            ' count timer
                        and      t1,#1                              ' mask out all but LSB
                        cmp      t1,#1 wz                           ' test if odd or even
                  if_z jmp      #serial_external                    ' odd so write to external ram
                        'put in hub ram line 0 =0, 2=1, 4=2 (y/2)*256 is the same as y*128
                        shl      t5,#7                              ' multiply by 128
                        add      t3,t5                              ' get hub location to store
serial_hub
                        rdbyte   t1,t2                              'get the first byte t1 = row number, multiply by 256 and add to t3
                        wrbyte   t1,t3                              ' store to hub ram
                        add      t2,#1                              ' increment source
                        add      t3,#1                              ' increment destination
                        djnz     t4,#serial_hub                    ' do 256 times            
                        jmp      #serial_finish                     ' set last byte to zero to say finished

serial_external        ' move data to external ram t5 contains row number - see pixels routine for the maths
                        rdbyte   t1,serial_buffer                   ' t1 is the row number 0-31 = block 0, 32-63=block1            
                        shr      t1,#5                              ' divide by 32
                        mov      latchcounter,t1                    ' set up latchcounter
                        call     #selectlatch                       ' select the correct latch
                        rdbyte   t3,serial_buffer                   ' get the row number again
                        and      t3,#%11111                         ' mask off so max is 31 as sram resets each block
                        sub      t3,#1                              ' subtract 1
                        shl      t3,#7                              ' multiply by 128 ie 256*(y-1)/2
                        mov      len,#256                           ' 256 bytes to write
                        mov      t4,serial_buffer                   ' serial buffer location
                        add      t4,#1                              ' add 1 so points to correct place
                        call     #wrblock                           ' write len bytes to sram t3 from hub t4
                        jmp      #serial_finish                     ' set last byte to zero to say finished reading

serial_finish           mov      t1,#0                              ' set value to non zero to say this has been read
                        wrbyte   t1,serial_buf_last                 ' store to the serial buffer
                        jmp      #main_loop_1
'rwo 1 =sram0,row 3=sram256, row 5=sram512, row33=sram0, block1 

finish     jmp #finish


' Initialised data
'address_mask    long %00000000_00000000_00011111_11111111                     ' 13 bits max 32 lines otherwise need to latch
oewrrd          long %00000000_11110000_00000000_00000000                     ' 23 = oe, 22=wr, 21=rd, latch high
oerd            long %00000000_01010000_00000000_00000000                     ' oe and rd low
oewr            long %00000000_00110000_00000000_00000000                     ' oe and wr low
dira_pins       long %00000000_11111111_11111111_00000000                     ' direction pins to be enabled                         
pin0to23        long %00000000_11111111_11111111_11111111                     ' for testing
zero            long %00000000_00000000_00000000_00000000                     ' for testing
'smallblock      long 1024
fivetwelve      long %00000000_00000000_00000010_00000000                       ' 512 for sram increment
blockmask       long %00000000_00010000_00000000_11111111                       ' mask for block select
latchlow        long %11111111_11101111_11111111_11111111                               ' mask for P20 latch                                                   
topram          long 28672

' Uninitialized data

status          res 1        ' status location in hub ram
t1              res 1        ' local variable
t2              res 1        ' local variable
t3              res 1        ' local variable
t4              res 1        ' local variable
t5              res 1        ' local variable
screen_buf      res 1        ' location of the screen buffer
_nextline       res 1        ' currrent line - updated by the video driver
sram_address    res 1        ' ram address value
hubaddr         res 1
len             res 1
current_line    res 1
pixels_x        res 1
localcounter    res 1
blockcounter    res 1
latchcounter    res 1
serial_buffer   res 1
serial_buf_last res 1
ramline         res 1 ' location of a single line in the top part of the screen memory that can be rewritten

        
        fit 496